package blastsearch;

import java.io.OutputStreamWriter;
import java.io.BufferedWriter;
import java.io.PrintWriter;

import java.util.Set;
import java.util.HashSet;
import java.util.Map;
import java.util.Iterator;
import java.util.LinkedList;

import org.biojava.bio.symbol.SymbolList;

import org.biojava.bio.search.SeqSimilaritySearcher;
import org.biojava.bio.search.SeqSimilaritySearchResult;
import org.biojava.bio.search.SequenceDBSearchResult;

import org.biojava.bio.program.sax.BlastLikeSAXParser;

import org.biojava.bio.program.ssbind.BlastLikeSearchBuilder;
import org.biojava.bio.program.ssbind.SeqSimilarityAdapter;

import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.db.SequenceDB;
import org.biojava.bio.seq.db.SimpleSequenceDBInstallation;

import org.biojava.bio.BioException;

import org.xml.sax.InputSource;

/**
 * This class performs BLAST-searches on a locally installed BLAST.
 *
 * Different BLAST-Programs (blastn, blastp, ...) must be specified via a
 * parameter, just like at the command line, since the interface doesn't
 * provide another way.
 *
 * How it works: Just call BLAST, pipe the output to the SAXParser, and connect
 * the Parser with the SearchBuilder.
 *
 * @author Joerg Ruedenauer
 * @version 0.1
 */
public class Blast implements SeqSimilaritySearcher {

  /**
   * Constructor. Constructs the DB-Set from the constant Array.
   */
  public Blast() {
    this.seqDBs = new SimpleSequenceDBInstallation();
    for (int i = 0; i < this.BLAST_DBS.length; i++) {
      try {
        this.seqDBs.addSequenceDB(SearchFactory.fastaProt2DB(
          BLAST_DBS[i]),  null);
      }
      catch (java.io.FileNotFoundException e) {
	  e.printStackTrace();
      }
    }
  }

  /**
   *  Return a set of all databases that can be searched with this sequence
   *  similarity searcher.
   *  @return Databases that can be searched.
   */
  public Set getSearchableDBs() {
    return this.seqDBs.getSequenceDBs();
  }

  /**
   * Using this sequence similarity searcher, search with the given sequence
   * against the given sequence database.
   *
   * This is the main method of Blast, and the only one with functionality.
   *
   * <b> The following is copied from the description of
   * org.biojava.search.SeqSimilaritySearcher:</b>
   * Using this sequence similarity searcher, search with the given sequence
   * against the given sequence database. It is a precondition that the sequence
   * and the sequence database be compatible with each other and with this
   * sequence similarity searcher, otherwise an IllegalArgumentException is
   * thrown.
   *
   * Particular implementations of a searcher will differ in the number and kind
   * of parameters that can be used to customize a search. All these parameters
   * must be passed as an argument to this method, i.e. these parameters are not
   * part of the state of this searcher.
   *
   * This method performs a synchronous search, i.e. it will block until the
   * search result becomes available.
   *
   * @param querySeq  the sequence with which to search. May not be null
   * otherwise an IllegalArgumentException is thrown.
   * @param the sequence database against which the similarity search will be
   * performed. May not be null otherwise an IllegalArgumentException is thrown.
   * Must also be an element of the set of searchable dbs returned by
   * getSearchableDBs().
   * @param  parameters that customize the search. Null is a legal
   * value for this argument and results in a default search being performed.
   * If this map contains keys and/or values that are not supported by BLAST, an
   * IllegalArgumentException is thrown. If no BLAST subprogram (blastn, blastp,
   * ...) is specified, blastp is used.
   * @return the sequence similarity search result that encapsulates the result
   * of the search. Never returns null.
   * @throws BioException  if the actual search fails for some reason. If,
   * however, the search can not even be started, because a precondition is not
   * met, an IllegalArgumentException is thrown.
   */
   public SeqSimilaritySearchResult search(SymbolList querySeq, SequenceDB db,
                                     Map searchParameters) throws BioException {
     if (querySeq == null || db == null || !(querySeq instanceof Sequence)) {
      throw new IllegalArgumentException();
     }
     // build the command line
     final String BLAST_PROGRAM = "blastall";
     String commandLine = SearchFactory.getBlastPath();
     commandLine += BLAST_PROGRAM + " ";
     if (searchParameters != null) {
       boolean programGiven = false;
       Iterator it = searchParameters.entrySet().iterator();
       while (it.hasNext()) {
         Map.Entry entry = (Map.Entry) it.next();
         if (!programGiven && entry.getKey().toString().equals("p")) {
          programGiven = true;
         }
         if (!entry.getKey().toString().equals("d")) {
           commandLine += " -" + entry.getKey() + " " + entry.getValue();
         }
         else if (entry.getValue().equals(db.getName())) {
	   System.out.println("Blast: note: param -d redundant");
         }
	 else {
           System.out.println("Blast: warning: param -d ignored");
	 }
       }
       if (!programGiven) {
         commandLine += " -p blastp";
       }
     }
     else {
      commandLine += " -p blastp";
     }
     commandLine += " -d " + db.getName();
     // initialize the connections
     java.util.LinkedList resultList = new java.util.LinkedList();
     BlastLikeSearchBuilder blsb =
       new BlastLikeSearchBuilder(resultList);
     blsb.setSubjectDBInstallation(this.seqDBs);
     blsb.setSubjectDB(db.getName());
     blsb.setQuerySeqHolder(db);
     String qSeqID = SearchFactory.createID(((Sequence)querySeq).getName());
     blsb.setQuerySeq(qSeqID);
     SeqSimilarityAdapter adapter = new SeqSimilarityAdapter();
     adapter.setSearchContentHandler(blsb);
     BlastLikeSAXParser blsp = new BlastLikeSAXParser();
     // Parser can only handle older blast versions if set to strict
     blsp.setModeLazy();
     blsp.setContentHandler(adapter);
     // start the BLAST search
     String[] envs = {"BLASTDB=" + SearchFactory.getBlastDBPath()};
     Process blast = null;
     try {	 
	 // System.err.println("Calling " + commandLine);
       blast = Runtime.getRuntime().exec(commandLine, envs);
       // complete the connections
       PrintWriter writer = new PrintWriter(new BufferedWriter(
         new OutputStreamWriter(blast.getOutputStream())));
       writer.println(">" + qSeqID);
       writer.println(querySeq.seqString());
       writer.println();
       // Print an 'End-Of-Input' - Symbol (Ctrl-D)
       writer.print('\u0004');
       writer.flush();
       writer.close();
       blsp.parse(new InputSource(blast.getInputStream()));
     }
     catch (java.io.IOException e1) {
       e1.printStackTrace();
       throw new BioException(e1.getMessage());
     }
     catch (org.xml.sax.SAXException e2) {
       e2.printStackTrace();
       throw new BioException(e2.getMessage());
     }
     // wait until the search has completed
     try {
       blast.waitFor();
     }
     catch (InterruptedException e3) {
     }
     // To-Do: Better Error-Handling through analysis of error-stream?
     if (blast.exitValue() != 0) {
       // probably a wrong argument - anyway, an error occurred
       throw new IllegalArgumentException();
     }
     // create Search Result
     SeqSimilaritySearchResult result = null;
     LinkedList hitList = new LinkedList();
     if (resultList.size() > 0) {
       Iterator it = resultList.iterator();
       while (it.hasNext()) {
	 SeqSimilaritySearchResult sssr = (SeqSimilaritySearchResult) it.next();
	 hitList.addAll(sssr.getHits());
	 // System.err.println(sssr.getHits().size() + " hits added.");
       }
       it = resultList.iterator();
       SequenceDBSearchResult sr = (SequenceDBSearchResult) it.next();
       result = new SequenceDBSearchResult(sr.getSequenceDB(),
					   sr.getSearchParameters(),
					   sr.getQuerySequence(), 
					   sr.getAnnotation(), 
					   hitList);
     }
     return result;
   }

  /**
   * The DBs that can be searched. To-Do: Read them from the path above. As yet,
   * only Protein DBs are supported. To-Do: More types.
   */
  private static final String[] BLAST_DBS = { "nr", "smalltestdb", 
    "lipases.dat", "lipsmalltest" };

  /**
   * The DBs that can be searched
   */
  private SimpleSequenceDBInstallation seqDBs;

}
