package blastsearch;

import org.biojava.bio.seq.Sequence;
import org.biojava.bio.seq.SequenceIterator;

import org.biojava.bio.seq.io.SeqIOTools;

import org.biojava.bio.seq.db.AbstractSequenceDB;

import java.util.Set;
import java.util.HashSet;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileNotFoundException;

/**
 * This class is a wrapper for FASTA Protein files so they can be used as
 * SequenceDBs. As yet, it doesn't support changes of the files. The file is
 * _not_ loaded into memory. The SequenceIterator thus operates directly on the
 * file; in fact, it's directly taken from org.biojava.seq.io.SeqIOTools. The
 * URNs of the sequences are used as IDs in the DB.
 *
 * @author Joerg Ruedenauer
 * @version 0.1
 */
public class FastaProt2SequenceDBWrapper extends AbstractSequenceDB {

  /**
   * Constructor.
   * @param name the name of the FASTA file.
   * @throws FileNotFoundException if the file doesn't exist.
   */
  public FastaProt2SequenceDBWrapper(String name) throws FileNotFoundException {
    super();
    this.name = name;
    // does the File exist?
    if (!(new File(SearchFactory.getBlastDBPath() + name)).exists()) {
      throw new FileNotFoundException("File " + name + " doesn't exist!");
    }
  }

  /**
   * Returns a sequence of the file. This could take long if the whole file
   * must be searched (flat file, sequential search!)
   * @see org.biojava.seq.db.SequenceDBLite.getSequence(String)
   */
  public Sequence getSequence(String id) throws
    org.biojava.bio.seq.db.IllegalIDException, org.biojava.bio.BioException {
    if (this.IDs != null) {
      // check the Set first, could avoid the search
      if (!this.IDs.contains(id)) {
        throw new org.biojava.bio.seq.db.IllegalIDException(
          "No Sequence with ID " + id + " in DB " + this.name);
      }
    }
    // search via Iterator
    SequenceIterator it = this.sequenceIterator();
    while (it.hasNext()) {
      try {
        Sequence s = it.nextSequence();
        if (SearchFactory.createID(s.getName()).equals(id)) {
          return s;
        }
      } catch (org.biojava.bio.BioException e) {
          e.printStackTrace();
          throw(e);
      }
    }
    throw new org.biojava.bio.seq.db.IllegalIDException(
      "No Sequence with ID " + id + " in DB " + this.name);
  }

  /**
   * Returns the IDs in the file. The first time it's called, the whole file
   * must be searched (could be slow).
   * @see org.biojava.seq.db.SequenceDB.ids()
   */
  public Set ids() {
    // Lazy creation of the set - only when it's needed, since the whole file
    // must be searched.
    if (this.IDs == null) {
      this.IDs = new HashSet();
      SequenceIterator it = this.sequenceIterator();
      while (it.hasNext()) {
        try {
          this.IDs.add(SearchFactory.createID(it.nextSequence().getName()));
        }
        catch (org.biojava.bio.BioException e) {
          // Something went wrong reading the file. But perhaps it was only
          // one sequence, so let's just ignore it and try to move on.
          e.printStackTrace();
        }
      }
    }
    // return the stored set
    return this.IDs;
  }

  /**
   * Returns the name of the DB.
   * @see org.biojava.seq.db.SequenceDBLite.getName()
   */
  public String getName() {
    return this.name;
  }

  /**
   * Adds a sequence to the DB (not yet allowed).
   * @see org.biojava.seq.db.SequenceDBLite.addSequence()
   */
  public void addSequence(Sequence seq) throws
  org.biojava.bio.BioException, org.biojava.utils.ChangeVetoException {
    throw new org.biojava.utils.ChangeVetoException(
      "Not yet supported by FastaProt2SequenceDBWrapper!");
  }

  /**
   * Removes a sequence from the DB (not yet allowed).
   * @see org.biojava.seq.db.SequenceDBLite.removeSequence()
   */
  public void removeSequence(Sequence seq) throws
  org.biojava.bio.BioException, org.biojava.utils.ChangeVetoException {
    throw new org.biojava.utils.ChangeVetoException(
      "Not yet supported by FastaProt2SequenceDBWrapper!");
  }

  /**
   * Returns a SequenceIterator.
   * @see org.biojava.seq.db.SequenceDB.sequenceIterator()
   */
  public SequenceIterator sequenceIterator() {
    try {
      // always use a new FileReader!
      return SeqIOTools.readFastaProtein(new BufferedReader(
        new FileReader(SearchFactory.getBlastDBPath() + this.name)));
    } catch (FileNotFoundException e ) {
      // should not occur... at least, the file was there then the instance
      // was constructed. Unfortunately, SequenceDB doesn't define exceptions
      // to be thrown by sequenceIterator().
      e.printStackTrace();
      return null;
    }
  }

  /**
   * Stores a Set of the IDs in the DB.
   */
  private Set IDs = null;

  /**
   * Stores the name of the DB (also the filename)
   */
  private String name;

}
